/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include "dash_exec.h"


#ifndef DS_TRACE_IF
# define DS_TRACE_IF  (z_mpi_rank == -1)
#endif


#ifdef WITH_SYMMETRIC

static void sym_disjoint_blocks(ds_exec_t *exec, int low0, int high0, int low1, int high1)
{
  int i, j;
  int s0, s1;


  s0 = high0 - low0;
  s1 = high1 - low1;

/*#define SEQ*/

  if (low0 <= exec->ds->comm_rank && exec->ds->comm_rank < high0)
  {
    for (i = 0; i < s1; ++i)
    {
#ifdef SEQ
      j = low1 + i;
#else
      j = low1 + ((i + (exec->ds->comm_rank - low0)) % s1);
#endif
      if (exec->ds->syms.counts[j] > 0)
        exec->sendrecv_replace(exec, j, exec->ds->syms.exec_id, exec->ds->syms.buf_ids[j], exec->ds->syms.displs[j], exec->ds->syms.counts[j]);
    }
  }
  
  if (low1 <= exec->ds->comm_rank && exec->ds->comm_rank < high1)
  {
    for (i = 0; i < s0; ++i)
    {
#ifdef SEQ
      j = low0 + i;
#else
      j = low0 + ((s0 - i + (exec->ds->comm_rank - low1)) % s0);
#endif
      if (exec->ds->syms.counts[j] > 0)
        exec->sendrecv_replace(exec, j, exec->ds->syms.exec_id, exec->ds->syms.buf_ids[j], exec->ds->syms.displs[j], exec->ds->syms.counts[j]);
    }
  }

#undef SEQ
}


dsint_t ds_exec_make_sym(ds_exec_t *exec) /* ds_func ds_exec_make_sym */
{
  Z_TRACE_IF(DS_TRACE_IF, "START");

  int l, h, b, m, o;


  if (exec->ds->syms.counts[exec->ds->comm_rank] > 0)
    exec->sendrecv_replace(exec, exec->ds->comm_rank, exec->ds->syms.exec_id,
      exec->ds->syms.buf_ids[exec->ds->comm_rank],
      exec->ds->syms.displs[exec->ds->comm_rank],
      exec->ds->syms.counts[exec->ds->comm_rank]);

  l = 1;
  h = exec->ds->comm_size;
  b = 0;

  while (l < h)
  {
    m = (l + h) / 2;
    o = (l + h + 1) % 2;

    sym_disjoint_blocks(exec, b, b + (h - m), m - o, h);

    if (exec->ds->comm_rank < m - o) h = m - o;
    else
    {
      l = m + (1 - o);
      b += (h - m);
    }
  }

  Z_TRACE_IF(DS_TRACE_IF, "END");

  return 0;
}

#endif


dsint_t ds_exec_make(ds_exec_t *exec) /* ds_func ds_exec_make */
{
#ifdef WITH_SYMMETRIC
  if (exec->ds->syms.execute)
  {
    ds_exec_make_sym(exec);
    exec->ds->syms.execute = 0;
  }
#endif

  if (exec->ds->sends.execute || exec->ds->recvs.execute || exec->ds->locals.execute)
  {
    exec->make(exec);
    exec->ds->sends.execute = exec->ds->recvs.execute = exec->ds->locals.execute = 0;
  }

  return 0;
}
