/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


/* sl_macro MEAS_TRACE_IF */

#include "sl_common.h"

#include "spec_core.h"


/* sl_ifdef SL_USE_MPI sl_context CONTEXT_BEGIN meas */
const double default_meas_t[2] = { 0 };     /* sl_global sl_context sl_var default_meas_t */
const slint_t default_meas_max_nprocs = 8;  /* sl_global sl_context sl_var default_meas_max_nprocs */
const slint_t default_meas_packed = 0;      /* sl_global sl_context sl_var default_meas_packed */
/* sl_endif sl_context CONTEXT_END meas */


#ifndef MEAS_TRACE_IF
# ifdef GLOBAL_TRACE_IF
#  define MEAS_TRACE_IF  GLOBAL_TRACE_IF
# else
#  define MEAS_TRACE_IF  (sl_mpi_rank == -1)
# endif
#endif


slint_t tproc_create(tproc_t *tproc, tproc_f *func, tprocs_f *func_s, tprocs_dup_f *func_s_dup) /* sl_proto, sl_func tproc_create */
{
  *tproc = z_alloc(1, sizeof(**tproc));

  spec_tproc_create(&(*tproc)->spec_tproc, func, func_s, func_s_dup);

  return 0;
}


slint_t tproc_destroy(tproc_t *tproc) /* sl_proto, sl_func tproc_destroy */
{
  spec_tproc_destroy(&(*tproc)->spec_tproc);

  z_free(*tproc);
  
  *tproc = NULL;

  return 0;
}


slint_t tproc_set_ext_tproc(tproc_t *tproc, tproc_count_f *count_db, tproc_count_f *count_ip, tproc_rearrange_db_f *rearrange_db, tproc_rearrange_ip_f *rearrange_ip) /* sl_proto, sl_func tproc_set_ext_tproc */
{
  spec_tproc_set_ext_tproc(&(*tproc)->spec_tproc, count_db, count_ip, rearrange_db, rearrange_ip);

  return 0;
}


slint_t tproc_set_ext_tprocs(tproc_t *tproc, tprocs_count_f *count_db, tprocs_count_f *count_ip, tprocs_rearrange_db_f *rearrange_db, tprocs_rearrange_ip_f *rearrange_ip) /* sl_proto, sl_func tproc_set_ext_tprocs */
{
  spec_tproc_set_ext_tprocs(&(*tproc)->spec_tproc, count_db, count_ip, rearrange_db, rearrange_ip);

  return 0;
}


slint_t tproc_set_ext_tprocs_dup(tproc_t *tproc, tprocs_dup_count_f *count_db, tprocs_dup_count_f *count_ip, tprocs_dup_rearrange_db_f *rearrange_db, tprocs_dup_rearrange_ip_f *rearrange_ip) /* sl_proto, sl_func tproc_set_ext_tprocs_dup */
{
  spec_tproc_set_ext_tprocs_dup(&(*tproc)->spec_tproc, count_db, count_ip, rearrange_db, rearrange_ip);

  return 0;
}


slint_t tproc_set_reset(tproc_t *tproc, tproc_reset_f reset) /* sl_proto, sl_func tproc_set_reset */
{
  spec_tproc_set_reset(&(*tproc)->spec_tproc, reset);

  return 0;
}


slint_t tproc_verify(tproc_t tproc, void *data, elements_t *s, int proc) /* sl_proto, sl_func tproc_verify */
{
  /* FIXME */

/*  slint_t i, j, n;
  int procs[SL_DEFCON(meas.max_nprocs)];


  for (i = 0; i < s->size; ++i)
  {
    n = tproc->tprocs_dup(s, i, data, procs, NULL);
    for (j = 0; j < n; ++j)
    if (procs[j] == proc) n = 0;

    if (n > 0) return 0;
  }*/

  return 1;
}


slint_t mpi_elements_alltoall_specific(elements_t *sin, elements_t *sout, elements_t *xs, tproc_t tproc, void *data, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_elements_alltoall_specific */
{
  slint_t r, original_mea_packed;

  original_mea_packed = SL_DEFCON(mea.packed); SL_DEFCON(mea.packed) = SL_DEFCON(meas.packed);

  if (sin == NULL || sout == NULL)
    r = spec_alltoallv_ip(sin, sout, xs, tproc->spec_tproc, data, size, rank, comm);
  else
    r = spec_alltoallv_db(sin, sout, xs, tproc->spec_tproc, data, size, rank, comm);

  SL_DEFCON(mea.packed) = original_mea_packed;
  
  return r;
}
