/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


/* sl_macro MEA_TRACE_IF */
/* sl_macro MEA_INPLACE_ALLTOALLV */

#include "sl_common.h"

#ifdef HAVE_ALLTOALLV_H
# include "alltoallv.h"
#endif

#include "dash_core.h"
#include "dash_sched_a2av.h"
#include "dash_sched_a2av_aux.h"
#include "dash_exec_sl.h"


/*#define MEA_INPLACE_ALLTOALLV*/

/* sl_ifdef SL_USE_MPI sl_context CONTEXT_BEGIN mea */
const slint_t default_mea_packed = 0;     /* sl_global sl_context sl_var default_mea_packed */
const slint_t default_mea_db_packed = 0;  /* sl_global sl_context sl_var default_mea_db_packed */
const slint_t default_mea_ip_packed = 0;  /* sl_global sl_context sl_var default_mea_ip_packed */
/* sl_endif sl_context CONTEXT_END mea */

#ifndef MEA_TRACE_IF
# ifdef GLOBAL_TRACE_IF
#  define MEA_TRACE_IF  GLOBAL_TRACE_IF
# else
#  define MEA_TRACE_IF  (SL_PROC_RANK == -1)
# endif
#endif


static slint_t _mpi_elements_alltoallv_db_packed(elements_t *sbuf, int *scounts, int *sdispls, elements_t *rbuf, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm, slint_t *lbs, slint_t *extents)
{
  slint_t i, _lbs[2] = { 0, 0 }, _extents[2] = { -1, -1 };
  slint_t local_exit, global_exit;
  packed_elements_t spackbuf, rpackbuf;
  MPI_Datatype packed_type;


  if (lbs == NULL) lbs = _lbs;
  if (extents == NULL) extents = _extents;

  if (extents[0] < 0) get_displcounts_extent(size, sdispls, scounts, &lbs[0], &extents[0]);
  if (extents[1] < 0) get_displcounts_extent(size, rdispls, rcounts, &lbs[1], &extents[1]);

  if (elements_alloc_packed(&spackbuf, extents[0]) != 0 || elements_alloc_packed(&rpackbuf, extents[1]) != 0)
  {
    elements_free_packed(&spackbuf);
    elements_free_packed(&rpackbuf);
    local_exit = 1;
  
  } else local_exit = 0;

  MPI_Allreduce(&local_exit, &global_exit, 1, int_mpi_datatype, MPI_SUM, comm);

  Z_TRACE_IF(MEA_TRACE_IF, "local_exit: %" slint_fmt ", global_exit: %" slint_fmt, local_exit, global_exit);

  if (global_exit) return -1;

  pelem_add(&spackbuf, -lbs[0]);
  pelem_add(&rpackbuf, -lbs[1]);

  mpi_elements_packed_datatype_create(&packed_type, 0);

  for (i = 0; i < size; ++i) elem_npack_at(sbuf, sdispls[i], &spackbuf, sdispls[i], scounts[i]);

  MPI_Alltoallv(spackbuf.elements, scounts, sdispls, packed_type, rpackbuf.elements, rcounts, rdispls, packed_type, comm);

  for (i = 0; i < size; ++i) pelem_nunpack_at(&rpackbuf, rdispls[i], rbuf, rdispls[i], rcounts[i]);

  mpi_elements_packed_datatype_destroy(&packed_type);

  pelem_add(&spackbuf, lbs[0]);
  pelem_add(&rpackbuf, lbs[1]);

  elements_free_packed(&spackbuf);
  elements_free_packed(&rpackbuf);

  return 0;
}


slint_t mpi_elements_alltoallv_db_packed(elements_t *sbuf, int *scounts, int *sdispls, elements_t *rbuf, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_elements_alltoallv_db_packed */
{
  return _mpi_elements_alltoallv_db_packed(sbuf, scounts, sdispls, rbuf, rcounts, rdispls, size, rank, comm, NULL, NULL);
}


slint_t mpi_elements_alltoallv_db(elements_t *sbuf, int *scounts, int *sdispls, elements_t *rbuf, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_elements_alltoallv_db */
{
  slint_t lbs[2] = { 0, 0 }, extents[2] = { -1, -1 };
  slint_t r = -1;

  Z_TRACE_IF(MEA_TRACE_IF, "packed: %" slint_fmt " or %" slint_fmt, SL_DEFCON(mea.packed), SL_DEFCON(mea.db_packed));

  if (SL_DEFCON(mea.packed) || SL_DEFCON(mea.db_packed))
    r = _mpi_elements_alltoallv_db_packed(sbuf, scounts, sdispls, rbuf, rcounts, rdispls, size, rank, comm, lbs, extents);

  if (r == 0) return 0;

#define xelem_call \
  MPI_Alltoallv(xelem_buf(sbuf), scounts, sdispls, xelem_mpi_datatype, xelem_buf(rbuf), rcounts, rdispls, xelem_mpi_datatype, comm);
#include "sl_xelem_call.h"

  return 0;
}


static slint_t _mpi_elements_alltoallv_ip_packed(elements_t *s, elements_t *sx, int *scounts, int *sdispls, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm, slint_t *lbs, slint_t *extents)
{
  slint_t i, _lbs[2] = { 0, 0 }, _extents[2] = { -1, -1 };
  slint_t local_exit, global_exit;
  void *block;
  slint_t blocksize;
  packed_elements_t packbuf, spackbuf, rpackbuf;
  MPI_Datatype packed_type;


  if (lbs == NULL) lbs = _lbs;
  if (extents == NULL) extents = _extents;

  if (extents[0] < 0) get_displcounts_extent(size, sdispls, scounts, &lbs[0], &extents[0]);
  if (extents[1] < 0) get_displcounts_extent(size, rdispls, rcounts, &lbs[1], &extents[1]);

  Z_TRACE_IF(MEA_TRACE_IF, "send: lb: %" slint_fmt ", extent: %" slint_fmt, lbs[0], extents[0]);
  Z_TRACE_IF(MEA_TRACE_IF, "receive: lb: %" slint_fmt ", extent: %" slint_fmt, lbs[1], extents[1]);

  if (sx)
  {
    elements_alloc_block(sx, &block, &blocksize, 8, -1);

    elements_alloc_packed_from_block(&packbuf, block, blocksize, 8, -1);

    Z_TRACE_IF(MEA_TRACE_IF, "alloc block size: %" slint_fmt ", alloc packed elements: %" slint_fmt, blocksize, packbuf.size);

    local_exit = (packbuf.size < extents[0] + extents[1])?1:0;

  } else local_exit = 1;

  MPI_Allreduce(&local_exit, &global_exit, 1, int_mpi_datatype, MPI_SUM, comm);

  Z_TRACE_IF(MEA_TRACE_IF, "local_exit: %" slint_fmt ", global_exit: %" slint_fmt, local_exit, global_exit);

  if (global_exit) return -1;

  pelem_assign(&packbuf, &spackbuf);
  pelem_assign_at(&packbuf, extents[0], &rpackbuf);

  mpi_elements_packed_datatype_create(&packed_type, 0);

  pelem_add(&spackbuf, -lbs[0]);
  pelem_add(&rpackbuf, -lbs[1]);

  for (i = 0; i < size; ++i) elem_npack_at(s, sdispls[i], &spackbuf, sdispls[i], scounts[i]);

  MPI_Alltoallv(spackbuf.elements, scounts, sdispls, packed_type, rpackbuf.elements, rcounts, rdispls, packed_type, comm);

  for (i = 0; i < size; ++i) pelem_nunpack_at(&rpackbuf, rdispls[i], s, rdispls[i], rcounts[i]);

  mpi_elements_packed_datatype_destroy(&packed_type);

  return 0;
}


slint_t mpi_elements_alltoallv_ip_packed(elements_t *s, elements_t *sx, int *scounts, int *sdispls, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_elements_alltoallv_ip_packed */
{
  return _mpi_elements_alltoallv_ip_packed(s, sx, scounts, sdispls, rcounts, rdispls, size, rank, comm, NULL, NULL);
}


static slint_t _mpi_elements_alltoallv_ip_double(elements_t *s, elements_t *sx, int *scounts, int *sdispls, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm, slint_t *lbs, slint_t *extents)
{
  slint_t i, _lbs[2] = { 0, 0 }, _extents[2] = { -1, -1 };
  slint_t local_exit, global_exit;
  elements_t _sx, _bx;
  void *block;
  slint_t blocksize;


  if (lbs == NULL) lbs = _lbs;
  if (extents == NULL) extents = _extents;

  if (extents[1] < 0) get_displcounts_extent(size, rdispls, rcounts, &lbs[1], &extents[1]);

  Z_TRACE_IF(MEA_TRACE_IF, "receive: lb: %" slint_fmt ", extent: %" slint_fmt, lbs[1], extents[1]);

  if (sx)
  {
    elements_alloc_block(sx, &block, &blocksize, 8, -1);

    Z_TRACE_IF(MEA_TRACE_IF, "alloc block size: %" slint_fmt, blocksize);

    if (blocksize >= extents[1] * elem_get_max_byte())
    {
      elem_set_block(&_bx, block);
      elem_set_block_size(&_bx, blocksize);
      sx = &_bx;
    }
      
    if (elem_get_block(sx))
    {
      Z_TRACE_IF(MEA_TRACE_IF, "block mem: size: %" slint_fmt ", max byte per component: %" slint_fmt, elem_get_block_size(sx), elem_get_max_byte());

      local_exit = (elem_get_block_size(sx) < extents[1] * elem_get_max_byte())?1:0;

    } else
    {
      Z_TRACE_IF(MEA_TRACE_IF, "elements mem: size: %" slint_fmt, elem_get_max_size(sx));

      local_exit = (elem_get_max_size(sx) < extents[1])?1:0;
    }

  } else local_exit = 1;


  MPI_Allreduce(&local_exit, &global_exit, 1, int_mpi_datatype, MPI_SUM, comm);

  Z_TRACE_IF(MEA_TRACE_IF, "local_exit: %" slint_fmt ", global_exit: %" slint_fmt, local_exit, global_exit);

  if (global_exit) return -1;
  
  if (elem_get_block(sx))
  {
#define xelem_call \
  xelem_set(&_sx, elem_get_block(sx)); \
  xelem_add(&_sx, -lbs[1]); \
  MPI_Alltoallv(xelem_buf(s), scounts, sdispls, xelem_mpi_datatype, xelem_get(&_sx), rcounts, rdispls, xelem_mpi_datatype, comm); \
  for (i = 0; i < size; ++i) xelem_ncopy_at(&_sx, rdispls[i], s, rdispls[i], rcounts[i]);
#include "sl_xelem_call.h"

  } else
  {
    elem_assign_at(sx, -lbs[1], &_sx);

#define xelem_call \
  MPI_Alltoallv(xelem_buf(s), scounts, sdispls, xelem_mpi_datatype, xelem_buf(&_sx), rcounts, rdispls, xelem_mpi_datatype, comm); \
  for (i = 0; i < size; ++i) xelem_ncopy_at(&_sx, rdispls[i], s, rdispls[i], rcounts[i]);
#include "sl_xelem_call.h"
  }
  
  return 0;
}


slint_t mpi_elements_alltoallv_ip_double(elements_t *s, elements_t *sx, int *scounts, int *sdispls, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_elements_alltoallv_ip_double */
{
  return _mpi_elements_alltoallv_ip_double(s, sx, scounts, sdispls, rcounts, rdispls, size, rank, comm, NULL, NULL);
}


slint_t mpi_elements_alltoallv_ip_mpi(elements_t *s, elements_t *sx, int *scounts, int *sdispls, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_elements_alltoallv_ip_mpi */
{
#if defined(HAVE_ALLTOALLV_H)

#define xelem_call \
  alltoallv_inplace_ds_aux = xelem_buf(sx); \
  alltoallv_inplace_ds_aux_size = sx->size * xelem_byte; \
  alltoallv_inplace_ds(MPI_IN_PLACE, scounts, sdispls, xelem_mpi_datatype, s, rcounts, rdispls, xelem_mpi_datatype, comm);
#include "sl_xelem_call.h"

  alltoallv_inplace_ds_aux = NULL;
  alltoallv_inplace_ds_aux_size = 0;

  return 0;

#else

  return -1;

#endif
}


slint_t mpi_elements_alltoallv_ip_dash(elements_t *s, elements_t *sx, int *scounts, int *sdispls, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_elements_alltoallv_ip_dash */
{
  ds_t ds;
  ds_sched_t sched;
  ds_sched_a2av_aux_t sched_a2av_aux;
  ds_exec_t exec;

  slint_t aux_size = 1000;  
  slint_t aux_blocks = -8;

  elements_t _sx, _bx;


  if (sx == NULL)
  {
    sx = &_sx;
    elements_alloc(sx, aux_size, SLCM_ALL);

  } else if (elem_get_block(sx))
  {
    sx = &_bx;
    elements_alloc_from_block(&_bx, elem_get_block(sx), elem_get_block_size(sx), 8, -1, SLCM_ALL);
  }

  ds_exec_sl_create(&exec);

  dsint_t s_addr_id = ds_exec_sl_add_address(&exec, s);
  dsint_t sx_addr_id = ds_exec_sl_add_address(&exec, sx);

  ds_sched_a2av_create(&sched);

  dsint_t s_buf_id = ds_sched_add_buffer(&sched, s_addr_id);
  ds_sched_set_send(&sched, s_buf_id, size, scounts, sdispls, NULL, 0);
  ds_sched_set_recv(&sched, s_buf_id, size, rcounts, rdispls, NULL, 0);

  dsint_t sx_buf_id = ds_sched_add_buffer(&sched, sx_addr_id);
  ds_sched_set_aux(&sched, sx_buf_id, sx->size);

  if (aux_blocks < 0) aux_blocks = size / -aux_blocks;
  aux_blocks = z_minmax(1, aux_blocks, size);

  ds_aux_static_create(&sched_a2av_aux, sx->size, aux_blocks);

  ds_sched_a2av_set_aux(&sched, &sched_a2av_aux);

  ds_create(&ds, &sched, &exec);
  
  ds.comm = comm;
  ds.comm_size = size;
  ds.comm_rank = rank;

  ds_run(&ds);

  ds_destroy(&ds);

  ds_aux_static_destroy(&sched_a2av_aux);

  ds_sched_a2av_destroy(&sched);

  ds_exec_sl_destroy(&exec);

  if (sx == &_sx) elements_free(sx);

  return 0;
}


slint_t mpi_elements_alltoallv_ip(elements_t *s, elements_t *sx, int *scounts, int *sdispls, int *rcounts, int *rdispls, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_elements_alltoallv_ip */
{
  slint_t lbs[2] = { 0, 0 }, extents[2] = { -1, -1 };
  slint_t r = -1;

  Z_TRACE_IF(MEA_TRACE_IF, "packed: %" slint_fmt " or %" slint_fmt, SL_DEFCON(mea.packed), SL_DEFCON(mea.ip_packed));

  if (SL_DEFCON(mea.packed) || SL_DEFCON(mea.ip_packed))
    r = _mpi_elements_alltoallv_ip_packed(s, sx, scounts, sdispls, rcounts, rdispls, size, rank, comm, lbs, extents);

  if (r == 0) return 0;

  r = _mpi_elements_alltoallv_ip_double(s, sx, scounts, sdispls, rcounts, rdispls, size, rank, comm, lbs, extents);

  if (r == 0) return 0;

#if defined(MEA_INPLACE_ALLTOALLV) && defined(HAVE_ALLTOALLV_H)

  r = mpi_elements_alltoallv_ip_mpi(s, sx, scounts, sdispls, rcounts, rdispls, size, rank, comm);

#else

  r = mpi_elements_alltoallv_ip_dash(s, sx, scounts, sdispls, rcounts, rdispls, size, rank, comm);
  
#endif

  return r;
}
