/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


/* sl_macro MSS_TRACE_IF */
/* sl_macro MSS_VERIFY */

#include "sl_common.h"

/*#define MSS_VERIFY*/

/* sl_ifdef SL_USE_MPI sl_context CONTEXT_BEGIN mssp */
const double default_mssp_i_t[3] = { 0 };  /* sl_global sl_context sl_var default_mssp_i_t */
const double default_mssp_p_t[3] = { 0 };  /* sl_global sl_context sl_var default_mssp_p_t */
const double default_mssp_b_t[3] = { 0 };  /* sl_global sl_context sl_var default_mssp_b_t */

const slint_t default_mssp_sync = 0;    /* sl_global sl_context sl_var default_mssp_sync */
const slint_t default_mssp_i_sync = 0;  /* sl_global sl_context sl_var default_mssp_i_sync */
const slint_t default_mssp_p_sync = 0;  /* sl_global sl_context sl_var default_mssp_p_sync */
const slint_t default_mssp_b_sync = 0;  /* sl_global sl_context sl_var default_mssp_b_sync */

const slint_t default_mssp_back_packed = 0;  /* sl_global sl_context sl_var default_mssp_back_packed */
/* sl_endif sl_context CONTEXT_END mssp */


#ifndef MSS_TRACE_IF
# ifdef GLOBAL_TRACE_IF
#  define MSS_TRACE_IF  GLOBAL_TRACE_IF
# else
#  define MSS_TRACE_IF  (sl_mpi_rank == -1)
# endif
#endif


static inline slint_t key2proc(slpkey_t k, slint_t nmm, slpkey_t *mm)
{
  slint_t l = 0;
  slint_t h = nmm - 1;
  slint_t m = (l + h) / 2;

  while (l < h)
  {
    if (sl_key_pure_cmp_le(k, mm[2 * m + 1])) h = m;
    else l = m + 1;

    m = (l + h) / 2;
  }

  return m;
}


#ifdef key_integer

slint_t mpi_sort_insert_radix(elements_t *s0, elements_t *s1, elements_t *xs, slpkey_t *mmkeys, slint_t rhigh, slint_t rlow, slint_t rwidth, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_sort_insert_radix */
{
  slint_t i, j;

  slpkey_t all_mmkeys[2 * size];
  
  int scounts[size], sdispls[size], rcounts[size], rdispls[size];


  if (SL_DEFCON(mssp.sync) || SL_DEFCON(mssp.i_sync)) MPI_Barrier(comm);

  /* local rearrange */
  SL_DEFCON(mssp.i_t)[0] = z_time_get_s();

  MPI_Allgather(mmkeys, 2, pkey_mpi_datatype, all_mmkeys, 2, pkey_mpi_datatype, comm);
  
  for (i = 0; i < size; ++i) scounts[i] = 0;

  for (i = 0; i < s0->size; ++i)
  {
    j = key2proc(key_purify(s0->keys[i]), size, all_mmkeys);
    ++scounts[j];
  }

  sdispls[0] = 0;
  for (i = 1; i < size; ++i) sdispls[i] = sdispls[i - 1] + scounts[i - 1];
  
  for (i = 0; i < s0->size; ++i)
  {
    j = key2proc(key_purify(s0->keys[i]), size, all_mmkeys);
    elem_copy_at(s0, i, s1, sdispls[j]);
    ++sdispls[j];
  }

  if (SL_DEFCON(mssp.sync) || SL_DEFCON(mssp.i_sync)) MPI_Barrier(comm);
  SL_DEFCON(mssp.i_t)[0] = z_time_get_s() - SL_DEFCON(mssp.i_t)[0];


  /* all-to-all */
  SL_DEFCON(mssp.i_t)[1] = z_time_get_s();

  sl_MPI_Alltoall_int(scounts, 1, rcounts, 1, comm, size, rank);

  sdispls[0] = rdispls[0] = 0;
  for (i = 1; i < size; ++i)
  {
    sdispls[i] = sdispls[i - 1] + scounts[i - 1];
    rdispls[i] = rdispls[i - 1] + rcounts[i - 1];
  }

  s0->size = s1->size = rdispls[size - 1] + rcounts[size - 1];

  mpi_elements_alltoallv_db(s1, scounts, sdispls, s0, rcounts, rdispls, size, rank, comm);

  if (SL_DEFCON(mssp.sync) || SL_DEFCON(mssp.i_sync)) MPI_Barrier(comm);
  SL_DEFCON(mssp.i_t)[1] = z_time_get_s() - SL_DEFCON(mssp.i_t)[1];


  /* local sort */
  SL_DEFCON(mssp.i_t)[2] = z_time_get_s();

  sort_radix(s0, s1, rhigh, rlow, -1);

  if (SL_DEFCON(mssp.sync) || SL_DEFCON(mssp.i_sync)) MPI_Barrier(comm);
  SL_DEFCON(mssp.i_t)[2] = z_time_get_s() - SL_DEFCON(mssp.i_t)[2];

  return 0;
}


slint_t mpi_sort_presorted_radix(elements_t *s0, elements_t *s1, elements_t *xs, slint_t merge_type, slint_t rhigh, slint_t rlow, slint_t rwidth, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_sort_presorted_radix */
{
  merge2x_f m2x;


  if (SL_DEFCON(mssp.sync) || SL_DEFCON(mssp.p_sync)) MPI_Barrier(comm);

  /* local sort */
  SL_DEFCON(mssp.p_t)[0] = z_time_get_s();

  sort_radix(s0, xs, rhigh, rlow, -1);

  if (SL_DEFCON(mssp.sync) || SL_DEFCON(mssp.p_sync)) MPI_Barrier(comm);
  SL_DEFCON(mssp.p_t)[0] = z_time_get_s() - SL_DEFCON(mssp.p_t)[0];


  /* merge sorted */
  SL_DEFCON(mssp.p_t)[1] = z_time_get_s();

  switch (merge_type)
  {
    case 0:
      m2x = merge2_basic_straight_01_x;
      break;
    case 1:
      m2x = merge2_compo_tridgell;
      break;
    case 2:
      m2x = merge2_compo_hula;
      break;
    default:
      m2x = merge2_memory_adaptive;
      break;
  }

  mpi_mergek_sorted(s0, m2x, xs, size, rank, comm);

  if (SL_DEFCON(mssp.sync) || SL_DEFCON(mssp.p_sync)) MPI_Barrier(comm);
  SL_DEFCON(mssp.p_t)[1] = z_time_get_s() - SL_DEFCON(mssp.p_t)[1];

  return 0;
}

#endif


typedef struct _tproc_sort_back_data
{
  slint_t nmm;
  slpkey_t *mm;

} tproc_sort_back_data;


int tproc_sort_back(elements_t *s, slint_t x, void *data) /* sl_func tproc_sort_back */
{
  tproc_sort_back_data *d = data;
  int l, h, m;


  if (key_purify(s->keys[x]) < 0) return -1;

  l = 0;
  h = d->nmm - 1;
  m = (l + h) / 2;

  while (l < h)
  {
    if (key_pure_cmp_le(key_purify(s->keys[x]), d->mm[2 * m + 1])) h = m;
    else l = m + 1;

    m = (l + h) / 2;
  }

  return m;
}


slint_t tprocs_sort_back(elements_t *s, slint_t x, void *data, int *procs) /* sl_func tprocs_sort_back */
{
  tproc_sort_back_data *d = data;
  int l, h, m;


  if (x < 0) return 1;

  if (key_purify(s->keys[x]) < 0) return 0;

  l = 0;
  h = d->nmm - 1;
  m = (l + h) / 2;

  while (l < h)
  {
    if (key_pure_cmp_le(key_purify(s->keys[x]), d->mm[2 * m + 1])) h = m;
    else l = m + 1;

    m = (l + h) / 2;
  }

  *procs = m;

  return 1;
}


slint_t tprocs_dup_sort_back(elements_t *s, slint_t x, void *data, int *procs, elements_t *dups) /* sl_func tprocs_dup_sort_back */
{
  tproc_sort_back_data *d = data;
  int l, h, m;


  if (x < 0) return -1;

  if (key_purify(s->keys[x]) < 0) return 0;

  l = 0;
  h = d->nmm - 1;
  m = (l + h) / 2;

  while (l < h)
  {
    if (key_pure_cmp_le(key_purify(s->keys[x]), d->mm[2 * m + 1])) h = m;
    else l = m + 1;

    m = (l + h) / 2;
  }

  *procs = m;

  return 1;
}


SPEC_DEFINE_TPROC(sort_back, tproc_sort_back, static);


slint_t mpi_sort_back(elements_t *sin, elements_t *sout, elements_t *sx, slpkey_t *lh, slint_t ntotal, int size, int rank, MPI_Comm comm) /* sl_proto, sl_func mpi_sort_back */
{
  slint_t i, j, original_meas_packed;

  slpkey_t lhs[2 * size];

  tproc_t tproc;
  tproc_sort_back_data data;
  
  elements_t *perm_src, *perm_dst, *perm_tmp, _sx, _bx;


  if (SL_DEFCON(mssp.sync) || SL_DEFCON(mssp.b_sync)) MPI_Barrier(comm);
  SL_DEFCON(mssp.b_t)[0] = z_time_get_s();

  MPI_Allgather(lh, 2, pkey_mpi_datatype, lhs, 2, pkey_mpi_datatype, comm);

  SL_DEFCON(mssp.b_t)[0] = z_time_get_s() - SL_DEFCON(mssp.b_t)[0];

/*  elements_print_keys(sin);*/

  if (sx != NULL && elem_get_block(sx))
  {
    elements_alloc_from_block(&_bx, elem_get_block(sx), elem_get_block_size(sx), 8, -1, SLCM_ALL);

    Z_TRACE_IF(MSS_TRACE_IF, "block mem: size: %" slint_fmt ", max byte per component: %" slint_fmt ", #elements: %" slint_fmt, elem_get_block_size(sx), elem_get_max_byte(), _bx.size);
    
    if (_bx.size > 0) sx = &_bx;
  }

  /* all-to-all specific */
  SL_DEFCON(mssp.b_t)[1] = z_time_get_s();

  data.nmm = size;
  data.mm = lhs;

  tproc_create(&tproc, tproc_sort_back, NULL, NULL);

  tproc_set_ext_tproc(&tproc, SPEC_EXT_PARAM_TPROC(sort_back));

  original_meas_packed = SL_DEFCON(meas.packed); SL_DEFCON(meas.packed) = SL_DEFCON(mssp.back_packed);

  mpi_elements_alltoall_specific(sin, sout, sx, tproc, &data, size, rank, comm);

  SL_DEFCON(meas.packed) = original_meas_packed;

  tproc_destroy(&tproc);

  SL_DEFCON(mssp.b_t)[1] = z_time_get_s() - SL_DEFCON(mssp.b_t)[1];

/*  elements_print_keys(sout);*/

  /* local permute back */
  SL_DEFCON(mssp.b_t)[2] = z_time_get_s();

  if (sout != NULL) perm_src = sout;
  else perm_src = sin;

  if (sx != NULL && sx->max_size >= perm_src->size) perm_dst = sx;
  else perm_dst = perm_src;

  if (perm_dst == perm_src && sx == NULL)
  {
    elements_alloc(&_sx, 1, SLCM_ALL);
    perm_tmp = &_sx;

  } else perm_tmp = sx;

/*  printf("%d: perm: %p, %p, %p\n", rank, perm_src, perm_dst, perm_tmp);*/

  if (perm_src != perm_dst)
  {
    for (i = 0; i < perm_src->size; ++i)
    {
      j = key_purify(perm_src->keys[i]) - lh[0];
      elem_copy_at(perm_src, i, perm_dst, j);
    }
    perm_dst->size = perm_src->size;

  } else
  {
    for (i = 0; i < perm_src->size; ++i)
    {
      j = key_purify(perm_src->keys[i]) - lh[0];
      while (j != i)
      {
        elem_xchange_at(perm_src, i, perm_src, j, perm_tmp);
        j = key_purify(perm_src->keys[i]) - lh[0];
      }
    }
  }

  if (sout != NULL)
  {
    if (perm_dst != sout) elem_ncopy(perm_dst, sout, perm_dst->size);

  } else
  {
    if (perm_dst != sin) elem_ncopy(perm_dst, sin, perm_dst->size);
  }

  if (perm_tmp == &_sx) elements_free(&_sx);

  if (SL_DEFCON(mssp.sync) || SL_DEFCON(mssp.b_sync)) MPI_Barrier(comm);
  SL_DEFCON(mssp.b_t)[2] = z_time_get_s() - SL_DEFCON(mssp.b_t)[2];

  return 0;
}
