/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#include "spec_core.h"


/*#define SPEC_PRINT*/


/*
static int dummy_spec_tproc(spec_elem_buf_t b, spec_elem_index_t x, spec_tproc_data_t tproc_data)
{
  return SPEC_;
}

SPEC_DEFINE_TPROC(dummy, dummy_spec_tproc);

static int dummy_spec_tprocs(spec_elem_buf_t b, spec_elem_index_t x, spec_tproc_data_t tproc_data, int *procs)
{
  return 0;
}

SPEC_DEFINE_TPROCS(dummy, dummy_spec_tprocs);

static int dummy_spec_tprocs_dup(spec_elem_buf_t b, spec_elem_index_t x, spec_tproc_data_t tproc_data, int *procs, spec_elem_buf_t dups)
{
  return 0;
}

SPEC_DEFINE_TPROCS_DUP(dummy, dummy_spec_tprocs_dup);
*/


spint_t spec_alltoallv_db(spec_elem_t *sb, spec_elem_t *rb, spec_elem_t *xb, spec_tproc_t tproc, spec_tproc_data_t tproc_data, int size, int rank, MPI_Comm comm) /* sp_func spec_alltoallv_db */
{
  spint_t exit_code = SPEC_EXIT_SUCCESS;

  spint_t i;

  int *procs = NULL;
  int *scounts, *sdispls, *rcounts, *rdispls;

  spint_t tprocs_max;
  spec_elem_t _sd, *sd, _xb;

  spint_t stotal, rtotal;

  SPEC_DECLARE_TPROC_COUNT_DB
  SPEC_DECLARE_TPROCS_COUNT_DB
  SPEC_DECLARE_TPROCS_DUP_COUNT_DB
  SPEC_DECLARE_TPROC_REARRANGE_DB
  SPEC_DECLARE_TPROCS_REARRANGE_DB
  SPEC_DECLARE_TPROCS_DUP_REARRANGE_DB

#ifdef Z_PACK_TIMING
  double t[5];
#endif


  Z_TIMING_SYNC(comm); Z_TIMING_START(t[0]);

  scounts = z_alloc(4 * size, sizeof(int));
  sdispls = scounts + 1 * size;
  rcounts = scounts + 2 * size;
  rdispls = scounts + 3 * size;

  /* determine max. duplicates */

  tprocs_max = 0;

  if (tproc->tprocs) tprocs_max = tproc->tprocs(spec_elem_get_buf(sb), -1, tproc_data, NULL);
  else if (tproc->tprocs_dup) tprocs_max = tproc->tprocs_dup(spec_elem_get_buf(sb), -1, tproc_data, NULL, NULL);

  sd = &_sd;

  if (tprocs_max < 0)
  {
    tprocs_max *= -1;
    sd = NULL;
  }

  if (tprocs_max)
  {
    procs = z_alloca(tprocs_max, sizeof(int));

    if (tproc->tprocs_dup)
    {
      spec_elem_alloc_buf(&_sd, tprocs_max);
      spec_elem_copy_type(sb, &_sd);
    }
  }

#ifdef SPEC_PRINT
  printf("input\n");
  spec_print(&tproc, tproc_data, sb);
#endif

  /* reset tproc if necessary */
  if (tproc->reset) tproc->reset(tproc_data);

  /* make local counts */
  Z_TIMING_SYNC(comm); Z_TIMING_START(t[1]);

  for (i = 0; i < size; ++i) scounts[i] = 0;

  if (tproc->tproc)
  {
    if (tproc->tproc_count_db) tproc->tproc_count_db(sb, tproc_data, scounts);
    else SPEC_DO_TPROC_COUNT_DB(tproc->tproc, tproc_data, sb, scounts);

  } else if (tproc->tprocs)
  {
    if (tproc->tprocs_count_db) tproc->tprocs_count_db(sb, tproc_data, scounts, procs);
    else SPEC_DO_TPROCS_COUNT_DB(tproc->tprocs, tproc_data, sb, scounts, procs);

  } else if (tproc->tprocs_dup)
  {
    if (tproc->tprocs_dup_count_db) tproc->tprocs_dup_count_db(sb, tproc_data, scounts, procs);
    else SPEC_DO_TPROCS_DUP_COUNT_DB(tproc->tprocs_dup, tproc_data, sb, scounts, procs);
  }

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[1]);

#ifdef SPEC_PRINT
  printf("after count\n");
  spec_print(&tproc, tproc_data, sb);
#endif

  /* redistribute local counts */
  Z_TIMING_SYNC(comm); Z_TIMING_START(t[2]);

#ifdef HAVE_ZMPI_TOOLS_H
# ifdef SPEC_MPI_ALLTOALL_2STEP_THRESHOLD
  if (size >= SPEC_MPI_ALLTOALL_2STEP_THRESHOLD)
    ZMPI_Alltoall_2step_int(scounts, 1, MPI_INT, rcounts, 1, MPI_INT, comm);
  else
# endif
#endif
    MPI_Alltoall(scounts, 1, MPI_INT, rcounts, 1, MPI_INT, comm);

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[2]);

  sdispls[0] = rdispls[0] = 0;
  for (i = 1; i < size; ++i)
  {
    sdispls[i] = sdispls[i - 1] + scounts[i - 1];
    rdispls[i] = rdispls[i - 1] + rcounts[i - 1];
  }

  stotal = sdispls[size - 1] + scounts[size - 1];
  rtotal = rdispls[size - 1] + rcounts[size - 1];

  if (rtotal > spec_elem_get_nmax(rb))
  {
#ifdef spec_elem_alloc_rbuf
    if (spec_elem_alloc_rbuf(rb)) spec_elem_alloc_buf(rb, rtotal);
    else
#endif
    {
#ifdef SPEC_ERROR_FILE
      fprintf(SPEC_ERROR_FILE, "%d: spec_alltoallv_db: error: receive buffer too small (%" spint_fmt " vs. %" spint_fmt")\n", rank, (spint_t) spec_elem_get_nmax(rb), rtotal);
#endif
      exit_code = SPEC_EXIT_FAILED;
      goto free_and_exit;
    }
  }

  if (xb == NULL || spec_elem_get_nmax(xb) < stotal)
  {
    spec_elem_copy_type(sb, &_xb);
    spec_elem_alloc_buf(&_xb, stotal);

    xb = &_xb;
  }

  spec_elem_copy_type(sb, xb);

  /* reset tproc if necessary */
  if (tproc->reset) tproc->reset(tproc_data);

  /* local rearrange */
  Z_TIMING_SYNC(comm); Z_TIMING_START(t[3]);

  if (tproc->tproc)
  {
    if (tproc->tproc_rearrange_db) tproc->tproc_rearrange_db(sb, xb, tproc_data, sdispls);
    else SPEC_DO_TPROC_REARRANGE_DB(tproc->tproc, tproc_data, sb, xb, sdispls);

  } else if (tproc->tprocs)
  {
    if (tproc->tprocs_rearrange_db) tproc->tprocs_rearrange_db(sb, xb, tproc_data, sdispls, procs);
    else SPEC_DO_TPROCS_REARRANGE_DB(tproc->tprocs, tproc_data, sb, xb, sdispls, procs);

  } else if (tproc->tprocs_dup)
  {
    if (tproc->tprocs_dup_rearrange_db) tproc->tprocs_dup_rearrange_db(sb, xb, tproc_data, sdispls, procs, sd);
    else SPEC_DO_TPROCS_DUP_REARRANGE_DB(tproc->tprocs_dup, tproc_data, sb, xb, sdispls, procs, sd);
  }

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[3]);

#ifdef SPEC_PRINT
  printf("after rearrange\n");
  spec_print(&tproc, tproc_data, xb);
#endif

  /* alltoallv */
  Z_TIMING_SYNC(comm); Z_TIMING_START(t[4]);

  sdispls[0] = 0;
  for (i = 1; i < size; ++i) sdispls[i] = sdispls[i - 1] + scounts[i - 1];

  spec_elem_alltoallv_db(xb, scounts, sdispls, rb, rcounts, rdispls, size, rank, comm);

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[4]);

  spec_elem_set_n(rb, rtotal);

  if (xb == &_xb) spec_elem_free_buf(&_xb);

free_and_exit:

  if (tprocs_max)
  {
    z_freea(procs);

    if (tproc->tprocs_dup) spec_elem_free_buf(&_sd);
  }

  z_free(scounts);

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[0]);

  Z_TIMING_PRINT(0, __func__, sizeof(t) / sizeof(t[0]), t, rank);

  return exit_code;
}


spint_t spec_alltoallv_ip(spec_elem_t *sb, spec_elem_t *rb, spec_elem_t *xb, spec_tproc_t tproc, spec_tproc_data_t tproc_data, int size, int rank, MPI_Comm comm) /* sp_func spec_alltoallv_ip */
{
  spint_t exit_code = SPEC_EXIT_SUCCESS;

  spint_t i;

  int *procs = NULL;
  int *scounts, *sdispls, *rcounts, *rdispls;

  spint_t tprocs_max;
  spec_elem_t _sd, *sd, _xb, b, *tb;

  spint_t stotal, rtotal;

  spint_t local_alltoallv_inplace, global_alltoallv_inplace;

  SPEC_DECLARE_TPROC_COUNT_IP
  SPEC_DECLARE_TPROC_REARRANGE_IP
  SPEC_DECLARE_TPROC_REARRANGE_DB
  SPEC_DECLARE_TPROCS_COUNT_IP
  SPEC_DECLARE_TPROCS_REARRANGE_DB
  SPEC_DECLARE_TPROCS_REARRANGE_IP
  SPEC_DECLARE_TPROCS_DUP_COUNT_IP
  SPEC_DECLARE_TPROCS_DUP_REARRANGE_DB
  SPEC_DECLARE_TPROCS_DUP_REARRANGE_IP

#ifdef Z_PACK_TIMING
  double t[5];
#endif


  Z_TIMING_SYNC(comm); Z_TIMING_START(t[0]);

  scounts = z_alloc(4 * size, sizeof(int));
  sdispls = scounts + 1 * size;
  rcounts = scounts + 2 * size;
  rdispls = scounts + 3 * size;

  /* determine max. duplicates */

  tprocs_max = 0;

  if (tproc->tprocs) tprocs_max = tproc->tprocs(spec_elem_get_buf(sb), -1, tproc_data, NULL);
  else if (tproc->tprocs_dup) tprocs_max = tproc->tprocs_dup(spec_elem_get_buf(sb), -1, tproc_data, NULL, NULL);

  sd = &_sd;

  if (tprocs_max < 0)
  {
    tprocs_max *= -1;
    sd = NULL;
  }

  if (tprocs_max)
  {
    procs = z_alloca(tprocs_max, sizeof(int));

    if (tproc->tprocs_dup) spec_elem_alloc_buf(&_sd, tprocs_max);
  }

  spec_elem_set_n(&b, spec_elem_get_n(sb));
  spec_elem_set_nmax(&b, spec_elem_get_nmax(sb));
  spec_elem_copy_type(sb, &b);

  if (spec_elem_get_buf(sb))
    spec_elem_set_buf(&b, spec_elem_get_buf(sb));
  else
    spec_elem_set_buf(&b, spec_elem_get_buf(rb));

#ifdef SPEC_PRINT
  printf("input\n");
  spec_print(&tproc, tproc_data, &b);
#endif

  /* reset tproc if necessary */
  if (tproc->reset) tproc->reset(tproc_data);

  /* make local counts */
  Z_TIMING_SYNC(comm); Z_TIMING_START(t[1]);

  for (i = 0; i < size; ++i) scounts[i] = 0;

  if (tproc->tproc)
  {
    if (tproc->tproc_count_ip) tproc->tproc_count_ip(&b, tproc_data, scounts);
    else SPEC_DO_TPROC_COUNT_IP(tproc->tproc, tproc_data, &b, scounts);

  } else if (tproc->tprocs)
  {
    if (tproc->tprocs_count_ip) tproc->tprocs_count_ip(&b, tproc_data, scounts, procs);
    else SPEC_DO_TPROCS_COUNT_IP(tproc->tprocs, tproc_data, &b, scounts, procs);

  } else if (tproc->tprocs_dup)
  {
    if (tproc->tprocs_dup_count_ip) tproc->tprocs_dup_count_ip(&b, tproc_data, scounts, procs);
    else SPEC_DO_TPROCS_DUP_COUNT_IP(tproc->tprocs_dup, tproc_data, &b, scounts, procs);
  }

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[1]);

#ifdef SPEC_PRINT
  printf("after count\n");
  spec_print(&tproc, tproc_data, &b);
#endif

  /* redistribute local counts */
  Z_TIMING_SYNC(comm); Z_TIMING_START(t[2]);

#ifdef SPEC_MPI_ALLTOALL_2STEP_THRESHOLD
  if (size >= SPEC_MPI_ALLTOALL_2STEP_THRESHOLD)
    ZMPI_Alltoall_2step_int(scounts, 1, MPI_INT, rcounts, 1, MPI_INT, comm);
  else
#endif
  MPI_Alltoall(scounts, 1, MPI_INT, rcounts, 1, MPI_INT, comm);

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[2]);

  sdispls[0] = rdispls[0] = 0;
  for (i = 1; i < size; ++i)
  {
    sdispls[i] = sdispls[i - 1] + scounts[i - 1];
    rdispls[i] = rdispls[i - 1] + rcounts[i - 1];
  }

  stotal = sdispls[size - 1] + scounts[size - 1];
  rtotal = rdispls[size - 1] + rcounts[size - 1];

  if (rtotal > spec_elem_get_nmax(rb))
  {
#ifdef SPEC_ERROR_FILE
    fprintf(SPEC_ERROR_FILE, "%d: spec_alltoallv_ip: error: receive buffer too small (%" spint_fmt " vs. %" spint_fmt")\n", rank, rtotal, (spint_t) spec_elem_get_nmax(&b));
#endif
    exit_code = SPEC_EXIT_FAILED;
    goto free_and_exit;
  }

  if (xb == NULL)
  {
    spec_elem_copy_type(sb, &_xb);
    spec_elem_alloc_buf(&_xb, 1);

    xb = &_xb;
  }

  tb = NULL;
  if (spec_elem_get_nmax(xb) >= stotal) tb = xb;

  spec_elem_copy_type(sb, xb);

  /* reset tproc if necessary */
  if (tproc->reset) tproc->reset(tproc_data);

  /* local rearrange */
  Z_TIMING_SYNC(comm); Z_TIMING_START(t[3]);

  if (tproc->tproc)
  {
    if (tb)
    {
      if (tproc->tproc_rearrange_db) tproc->tproc_rearrange_db(&b, tb, tproc_data, sdispls);
      else SPEC_DO_TPROC_REARRANGE_DB(tproc->tproc, tproc_data, &b, tb, sdispls);

    } else
    {
      if (tproc->tproc_rearrange_ip) tproc->tproc_rearrange_ip(&b, xb, tproc_data, sdispls, scounts, size);
      else SPEC_DO_TPROC_REARRANGE_IP(tproc->tproc, tproc_data, &b, xb, sdispls, scounts, size);
    }

  } else if (tproc->tprocs)
  {
    if (tb)
    {
      if (tproc->tprocs_rearrange_db) tproc->tprocs_rearrange_db(&b, tb, tproc_data, sdispls, procs);
      else SPEC_DO_TPROCS_REARRANGE_DB(tproc->tprocs, tproc_data, &b, tb, sdispls, procs);

    } else
    {
      if (tproc->tprocs_rearrange_ip) tproc->tprocs_rearrange_ip(&b, xb, tproc_data, sdispls, scounts, size, procs);
      else SPEC_DO_TPROCS_REARRANGE_IP(tproc->tprocs, tproc_data, &b, xb, sdispls, scounts, size, procs);
    }

  } else if (tproc->tprocs_dup)
  {
    if (tb)
    {
      if (tproc->tprocs_dup_rearrange_db) tproc->tprocs_dup_rearrange_db(&b, tb, tproc_data, sdispls, procs, sd);
      else SPEC_DO_TPROCS_DUP_REARRANGE_DB(tproc->tprocs_dup, tproc_data, &b, tb, sdispls, procs, sd);

    } else
    {
      if (tproc->tprocs_dup_rearrange_ip) tproc->tprocs_dup_rearrange_ip(&b, xb, tproc_data, sdispls, scounts, size, procs, sd);
      else SPEC_DO_TPROCS_DUP_REARRANGE_IP(tproc->tprocs_dup, tproc_data, &b, xb, sdispls, scounts, size, procs, sd);
    }
  }

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[3]);

#ifdef SPEC_PRINT
  printf("after rearrange\n");
  spec_print(&tproc, tproc_data, &b);
#endif

  /* alltoallv */
  Z_TIMING_SYNC(comm); Z_TIMING_START(t[4]);

  sdispls[0] = 0;
  for (i = 1; i < size; ++i) sdispls[i] = sdispls[i - 1] + scounts[i - 1];

  if (tb) local_alltoallv_inplace = 0;
  else
  {
    if (spec_elem_get_nmax(xb) >= rtotal) local_alltoallv_inplace = 0;
    else local_alltoallv_inplace = 1;
  }

  MPI_Allreduce(&local_alltoallv_inplace, &global_alltoallv_inplace, 1, MPI_SPINT, MPI_SUM, comm);

  /* no process wants inplace? */
  if (global_alltoallv_inplace == 0)
  {
    /* double-buffered alltoallv  */
    if (tb) spec_elem_alltoallv_db(tb, scounts, sdispls, &b, rcounts, rdispls, size, rank, comm);
    else
    {
      spec_elem_alltoallv_db(&b, scounts, sdispls, xb, rcounts, rdispls, size, rank, comm);
      spec_elem_ncopy_at(xb, 0, &b, 0, rtotal);
    }

  } else
  {
#ifdef spec_elem_alltoallv_ip

    /* in-place alltoallv  */
    if (tb) spec_elem_ncopy_at(tb, 0, &b, 0, stotal);
    spec_elem_alltoallv_ip(&b, xb, scounts, sdispls, rcounts, rdispls, size, rank, comm);

#else

#ifdef SPEC_ERROR_FILE
    fprintf(SPEC_ERROR_FILE, "%d: spec_alltoallv_ip: error: required in-place MPI_Alltoallv support not available, use the ZMPI All-to-all In-place library!");
#endif
    exit_code = SPEC_EXIT_FAILED;
    goto free_and_exit;

#endif
  }

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[4]);

  if (spec_elem_get_buf(sb))
    spec_elem_set_n(sb, rtotal);
  else
    spec_elem_set_n(rb, rtotal);

  if (xb == &_xb) spec_elem_free_buf(&_xb);

free_and_exit:

  if (tprocs_max)
  {
    z_freea(procs);

    if (tproc->tprocs_dup) spec_elem_free_buf(&_sd);
  }

  z_free(scounts);

  Z_TIMING_SYNC(comm); Z_TIMING_STOP(t[0]);

  Z_TIMING_PRINT(0, __func__, sizeof(t) / sizeof(t[0]), t, rank);

  return exit_code;
}
