/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#include "spec_core.h"


static void spec_tproc_unset_tproc(spec_tproc_t *tproc)
{
  (*tproc)->tproc = NULL;
}


static void spec_tproc_unset_ext_tproc(spec_tproc_t *tproc)
{
  (*tproc)->tproc_count_db = NULL;
  (*tproc)->tproc_count_ip = NULL;
  (*tproc)->tproc_rearrange_db = NULL;
  (*tproc)->tproc_rearrange_ip = NULL;
}


static void spec_tproc_unset_tprocs(spec_tproc_t *tproc)
{
  (*tproc)->tprocs = NULL;
}


static void spec_tproc_unset_ext_tprocs(spec_tproc_t *tproc)
{
  (*tproc)->tprocs_count_db = NULL;
  (*tproc)->tprocs_count_ip = NULL;
  (*tproc)->tprocs_rearrange_db = NULL;
  (*tproc)->tprocs_rearrange_ip = NULL;
}


static void spec_tproc_unset_tprocs_dup(spec_tproc_t *tproc)
{
  (*tproc)->tprocs_dup = NULL;
}


static void spec_tproc_unset_ext_tprocs_dup(spec_tproc_t *tproc)
{
  (*tproc)->tprocs_dup_count_db = NULL;
  (*tproc)->tprocs_dup_count_ip = NULL;
  (*tproc)->tprocs_dup_rearrange_db = NULL;
  (*tproc)->tprocs_dup_rearrange_ip = NULL;
}


spint_t spec_tproc_create(spec_tproc_t *tproc, spec_tproc_f *func, spec_tprocs_f *func_s, spec_tprocs_dup_f *func_s_dup) /* sp_func spec_tproc_create */
{
  *tproc = z_alloc(1, sizeof(struct _spec_tproc_t));

  (*tproc)->tproc = func;
  spec_tproc_unset_ext_tproc(tproc);

  (*tproc)->tprocs = func_s;
  spec_tproc_unset_ext_tprocs(tproc);

  (*tproc)->tprocs_dup = func_s_dup;
  spec_tproc_unset_ext_tprocs_dup(tproc);

  (*tproc)->reset = NULL;

  return SPEC_EXIT_SUCCESS;
}


spint_t spec_tproc_destroy(spec_tproc_t *tproc) /* sp_func spec_tproc_destroy */
{
  z_free(*tproc);
  
  *tproc = NULL;

  return SPEC_EXIT_SUCCESS;
}


spint_t spec_tproc_set_tproc(spec_tproc_t *tproc, spec_tproc_f *func) /* sp_func spec_tproc_set_tproc */
{
  spec_tproc_unset_tproc(tproc);
  spec_tproc_unset_tprocs(tproc);
  spec_tproc_unset_tprocs_dup(tproc);

  (*tproc)->tproc = func;

  return SPEC_EXIT_SUCCESS;
}


spint_t spec_tproc_set_ext_tproc(spec_tproc_t *tproc, spec_tproc_count_f *func_count_db, spec_tproc_count_f *func_count_ip, spec_tproc_rearrange_db_f *func_rearrange_db, spec_tproc_rearrange_ip_f *func_rearrange_ip) /* sp_func spec_tproc_set_ext_tproc */
{
  (*tproc)->tproc_count_db = func_count_db;
  (*tproc)->tproc_count_ip = func_count_ip;
  (*tproc)->tproc_rearrange_db = func_rearrange_db;
  (*tproc)->tproc_rearrange_ip = func_rearrange_ip;

  return SPEC_EXIT_SUCCESS;
}


spint_t spec_tproc_set_tprocs(spec_tproc_t *tproc, spec_tprocs_f *func_s) /* sp_func spec_tproc_set_tprocs */
{
  spec_tproc_unset_tproc(tproc);
  spec_tproc_unset_tprocs(tproc);
  spec_tproc_unset_tprocs_dup(tproc);

  (*tproc)->tprocs = func_s;

  return SPEC_EXIT_SUCCESS;
}


spint_t spec_tproc_set_ext_tprocs(spec_tproc_t *tproc, spec_tprocs_count_f *func_count_db, spec_tprocs_count_f *func_count_ip, spec_tprocs_rearrange_db_f *func_rearrange_db, spec_tprocs_rearrange_ip_f *func_rearrange_ip) /* sp_func spec_tproc_set_ext_tprocs */
{
  (*tproc)->tprocs_count_db = func_count_db;
  (*tproc)->tprocs_count_ip = func_count_ip;
  (*tproc)->tprocs_rearrange_db = func_rearrange_db;
  (*tproc)->tprocs_rearrange_ip = func_rearrange_ip;

  return SPEC_EXIT_SUCCESS;
}


spint_t spec_tproc_set_tprocs_dup(spec_tproc_t *tproc, spec_tprocs_dup_f *func_s_dup) /* sp_func spec_tproc_set_tprocs_dup */
{
  spec_tproc_unset_tproc(tproc);
  spec_tproc_unset_tprocs(tproc);
  spec_tproc_unset_tprocs_dup(tproc);

  (*tproc)->tprocs_dup = func_s_dup;

  return SPEC_EXIT_SUCCESS;
}


spint_t spec_tproc_set_ext_tprocs_dup(spec_tproc_t *tproc, spec_tprocs_dup_count_f *func_count_db, spec_tprocs_dup_count_f *func_count_ip, spec_tprocs_dup_rearrange_db_f *func_rearrange_db, spec_tprocs_dup_rearrange_ip_f *func_rearrange_ip) /* sp_func spec_tproc_set_ext_tprocs_dup */
{
  (*tproc)->tprocs_dup_count_db = func_count_db;
  (*tproc)->tprocs_dup_count_ip = func_count_ip;
  (*tproc)->tprocs_dup_rearrange_db = func_rearrange_db;
  (*tproc)->tprocs_dup_rearrange_ip = func_rearrange_ip;

  return SPEC_EXIT_SUCCESS;
}


spint_t spec_tproc_set_reset(spec_tproc_t *tproc, spec_tproc_reset_f *reset) /* sp_func spec_tproc_set_reset */
{
  (*tproc)->reset = reset;
  
  return SPEC_EXIT_SUCCESS;
}


spint_t spec_print(spec_tproc_t *tproc, spec_tproc_data_t tproc_data, spec_elem_t *b) /* sp_func spec_print */
{
  spint_t i, j, n;
  int p, *procs = NULL;

  spint_t tprocs_max, tprocs_dup;
  spec_elem_t sd;
  
  
  tprocs_max = 0;
  tprocs_dup = 1;

  if ((*tproc)->tprocs) tprocs_max = (*tproc)->tprocs(spec_elem_get_buf(b), -1, tproc_data, NULL);
  else if ((*tproc)->tprocs_dup) tprocs_max = (*tproc)->tprocs_dup(spec_elem_get_buf(b), -1, tproc_data, NULL, NULL);

  if (tprocs_max < 0)
  {
    tprocs_max *= -1;
    tprocs_dup = 0;
  }

  if (tprocs_max)
  {
    procs = z_alloca(tprocs_max, sizeof(int));

    if ((*tproc)->tprocs_dup) spec_elem_alloc_buf(&sd, tprocs_max);
  }

  /* reset tproc if necessary */
  if ((*tproc)->reset) (*tproc)->reset(tproc_data);


  if ((*tproc)->tproc)
  {
    for (i = 0; i < spec_elem_get_n(b); ++i)
    {
      p = (*tproc)->tproc(spec_elem_get_buf(b), i, tproc_data);
      printf("%" spint_fmt ": %d\n", i, p);
    }

  } else if ((*tproc)->tprocs)
  {
    for (i = 0; i < spec_elem_get_n(b); ++i)
    {
      n = (*tproc)->tprocs(spec_elem_get_buf(b), i, tproc_data, procs);
      printf("%" spint_fmt ":", i);
      for (j = 0; j < n; ++j) printf(" %d", procs[j]);
      printf("\n");
    }

  } else if ((*tproc)->tprocs_dup)
  {
    for (i = 0; i < spec_elem_get_n(b); ++i)
    {
      n = (*tproc)->tprocs_dup(spec_elem_get_buf(b), i, tproc_data, procs, (tprocs_dup)?&sd:NULL);
      printf("%" spint_fmt ":", i);
      for (j = 0; j < n; ++j) printf(" %d", procs[j]);
      printf("\n");
    }
  }
  
  if (tprocs_max)
  {
    z_freea(procs);

    if ((*tproc)->tprocs_dup) spec_elem_free_buf(&sd);
  }

  return 0;
}
