/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include "sl_common.h"


#ifdef sl_key_integer

void demo_radix()
{
  slint_t n, i, j;
  slint_t nlow = 1, nhigh = 4*1024*1024;

  double t[5];
  slint_t o[5];
  
  elements_t s, s2, r, xs;

  sl_key_type_c kmin, kmax;
  
  slint_t rlow = 0;
  slint_t rhigh = 30;
  slint_t rwidth = -1;
  

  printf("demo_radix\n");

  elements_alloc(&s, nhigh, SLCM_ALL);
  elements_alloc(&s2, nhigh, SLCM_ALL);
  elements_alloc(&r, nhigh, SLCM_ALL);
  elements_alloc(&xs, nhigh, SLCM_ALL);

  for (n = 0; n < nhigh; n++) s.keys[n] = random();
  
  kmin = kmax = s.keys[0];

  for (n = 0; n < nhigh; n++)
  {
    if (key_cmp_lt(s.keys[n], kmin)) kmin = s.keys[n];
    if (key_cmp_gt(s.keys[n], kmax)) kmax = s.keys[n];
  }
  
  printf("kmin = %" sl_key_type_fmt " - kmax = %" sl_key_type_fmt " - RAND_MAX = %d\n", kmin, kmax, RAND_MAX);

#undef elements_touch
#define elements_touch(_s_)

  for (n = nlow; n <= nhigh; n *= 2)
  {
    elem_ncopy(&s, &s2, n);
    sort_radix_ip(&s2, NULL, -1, -1, -1);
    
/*    for (j = 0; j < (n * 1) / 10000; ++j) s2.keys[random() % n] = random();*/
/*    for (j = 0; j < 40; ++j) s2.keys[random() % n] = random();*/
    
    elem_ncopy(&s2, &r, n);
    r.size = n;
    elements_touch(&r);
    elements_touch(&xs);

    i = 0;
    t[i] = sl_ts_get_ms();
    sort_radix_db(&r, &xs, rhigh, rlow, rwidth);
    t[i] = sl_ts_get_ms() - t[i];
    o[i] = elements_validate_order(&r, 1);

    elem_ncopy(&s2, &r, n);
    r.size = n;
    elements_touch(&r);
    elements_touch(&xs);

    i = 1;
    t[i] = sl_ts_get_ms();
    sort_radix_ip(&r, &xs, rhigh, rlow, rwidth);
    t[i] = sl_ts_get_ms() - t[i];
    o[i] = elements_validate_order(&r, 1);

    elem_ncopy(&s2, &r, n);
    r.size = n;
    elements_touch(&r);
    elements_touch(&xs);

    i = 2;
    t[i] = sl_ts_get_ms();
    sort_radix(&r, &xs, rhigh, rlow, rwidth);
    t[i] = sl_ts_get_ms() - t[i];
    o[i] = elements_validate_order(&r, 1);

    elem_ncopy(&s2, &r, n);
    r.size = n;
    elements_touch(&r);

    i = 3;
    t[i] = sl_ts_get_ms();
    sort_radix_af(&r, NULL, rhigh, rlow, rwidth);
    t[i] = sl_ts_get_ms() - t[i];
    o[i] = elements_validate_order(&r, 1);

    elem_ncopy(&s2, &r, n);
    r.size = n;
    elements_touch(&r);

    i = 4;
    t[i] = sl_ts_get_ms();
    sort_radix_1bit(&r, NULL, rhigh, rlow);
    t[i] = sl_ts_get_ms() - t[i];
    o[i] = elements_validate_order(&r, 1);

    printf("%" sl_int_type_fmt "  %f%s  %f%s  %f%s  %f%s  %f%s\n", n, t[0], (o[0])?"!":"", t[1], (o[1])?"!":"", t[2], (o[2])?"!":"", t[3], (o[3])?"!":"", t[4], (o[4])?"!":"");
  }

  elements_free(&s);
  elements_free(&r);
}

#endif


int main(int argc, char *argv[])
{
  demo_radix();

  return 0;
}
