/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include "sl_common.h"


void demo_parallel_merge_based_sort()
{
  MPI_Comm comm = MPI_COMM_WORLD;
  int size, rank;

  slint_t n;
  slint_t nlow = 1, nhigh = 1*1024*1024;

  double t1, t2, t3;
  
  elements_t s, r;

#define BUFSIZE  1024
  char buf[BUFSIZE];


  SL_DEFCON(me.sendrecv_replace_mem) = buf;
  SL_DEFCON(me.sendrecv_replace_memsize) = BUFSIZE;

  MPI_Comm_size(comm, &size);
  MPI_Comm_rank(comm, &rank);
  
  SL_DEFCON(mpi.rank) = rank;

  if (rank == 0) printf("demo_parallel_merge_based_sort: #processes: %d\n", size);

  elements_alloc(&s, nhigh, SLCM_ALL);
  elements_alloc(&r, nhigh, SLCM_ALL);

  for (n = 0; n < nhigh; n++) s.keys[n] = random();

  for (n = nlow; n <= nhigh; n *= 2)
  {
    elem_ncopy(&s, &r, n);
    r.size = n;
    elements_touch(&r);

    MPI_Barrier(comm);
    t1 = z_time_wtime();

    t2 = z_time_wtime();
#ifdef sl_key_integer
    sort_radix(&r, NULL, 30, 0, -1);
#else
    sort_quick(&r, NULL);
#endif
    t2 = z_time_wtime() - t2;

    t3 = z_time_wtime();
    mpi_mergek(&r, sn_batcher, NULL, merge2_compo_hula, NULL, size, rank, comm);
    t3 = z_time_wtime() - t3;

    MPI_Barrier(comm);
    t1 = z_time_wtime() - t1;

    if (rank == 0) printf("%" sl_int_type_fmt "  %f  %f  %f\n", n, t1, t2, t3);
  }
  
  elements_free(&s);
  elements_free(&r);

/*  printf("%d: rti_mem: nalloc: %d, nfree: %d\n", rank, rti_env.mem.nalloc, rti_env.mem.nfree);
  printf("%d: rti_mem: max: %d, cur: %d, cur_max: %d\n", rank, rti_env.mem.max, rti_env.mem.cur, rti_env.mem.cur_max);*/
}


int main(int argc, char *argv[])
{
  MPI_Init(&argc, &argv);

  mpi_datatypes_init();

  demo_parallel_merge_based_sort();

  mpi_datatypes_release();

  MPI_Finalize();

  return 0;
}
