/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include "sl_common.h"


tune_f seq_tunes[] = {
  tune_sort_radix_width,
  tune_sort_radix_threshold,
  NULL
};


void tune_auto_intro(int argc, char *argv[])
{
  slint_t i;


  fprintf(TUNE_RESULT_STREAM,
          "/*\n"
          "   This file was automatically generated by the automatic tuning support!\n"
          "   Call:");

  for (i = 0; i < argc; ++i) fprintf(TUNE_RESULT_STREAM, " %s", argv[i]);
  
  fprintf(TUNE_RESULT_STREAM,
          "\n"
          "*/\n"
          "\n"
          "#ifndef __SL_TUNE_AUTO_H__\n"
          "#define __SL_TUNE_AUTO_H__\n"
          "\n"
          "\n");
}


void tune_auto_outro()
{
  fprintf(TUNE_RESULT_STREAM,
          "\n"
          "#endif /* __SL_TUNE_AUTO_H__ */\n");
}


void tune_auto_begin(const char *name)
{
  fprintf(TUNE_PROGRESS_STREAM, "%s: BEGIN\n", name);
}


void tune_auto_end(const char *name)
{
  fprintf(TUNE_PROGRESS_STREAM, "%s: END\n", name);
}


void tune_auto_output_result(const char *key, const char *val)
{
  fprintf(TUNE_RESULT_STREAM,
          "#undef %s\n"
          "#define %s  %s\n"
          "\n",
          key, key, val);
}


void tune_auto_output_result_slint(const char *key, slint_t val)
{
  char s[64];

  snprintf(s, 64, "%" slint_fmt, val);

  tune_auto_output_result(key, s);
}


int main(int argc, char *argv[])
{
  slint_t i;


  tune_auto_intro(argc, argv);

  for (i = 0; i < sizeof(seq_tunes) / sizeof(tune_f) && seq_tunes[i] != NULL; ++i)
  {
    seq_tunes[i](NULL);
  }

  tune_auto_outro();
  
  return 0;
}
