/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


/* sl_macro TC_TRACE_IF */


#include "sl_common.h"


#ifndef TC_TRACE_IF
# ifdef GLOBAL_TRACE_IF
#  define TC_TRACE_IF  GLOBAL_TRACE_IF
# else
#  define TC_TRACE_IF  (sl_mpi_rank != -1)
# endif
#endif


static double det3x3(double *v0, double *v1, double *v2)
{
  return v0[0] * v1[1] * v2[2]
       + v1[0] * v2[1] * v0[2]
       + v2[0] * v0[1] * v1[2]
       - v0[2] * v1[1] * v2[0]
       - v1[2] * v2[1] * v0[0]
       - v2[2] * v0[1] * v1[0];
}


void interpolate_poly2(slint_t n, double *x, double *y, double *a, double *b, double *c)
{
  slint_t i;
  double A[9], v[3], det;
  
  
  for (i = 0; i < 9; ++i) A[i] = 0;

  v[0] = v[1] = v[2] = 0;
  
  for (i = 0; i < n; ++i)
  {
    A[8] += 1;
    A[7] += x[i];
    A[6] += x[i] * x[i];
    A[5] = A[7];
    A[4] = A[6];
    A[3] += x[i] * x[i] * x[i];
    A[2] = A[4];
    A[1] = A[3];
    A[0] += x[i] * x[i] * x[i] * x[i];

    v[0] += x[i] * x[i] * y[i];
    v[1] += x[i] * y[i];
    v[2] += y[i];
  }

  Z_ERROR_IF(TC_TRACE_IF, "%f  %f  %f ... %f", A[0], A[3], A[6], v[0]);
  Z_ERROR_IF(TC_TRACE_IF, "%f  %f  %f ... %f", A[1], A[4], A[7], v[0]);
  Z_ERROR_IF(TC_TRACE_IF, "%f  %f  %f ... %f", A[2], A[5], A[8], v[0]);

  det = det3x3(&A[0 * 3], &A[1 * 3], &A[2 * 3]);
  *a = det3x3(v, &A[1 * 3], &A[2 * 3]) / det;
  *b = det3x3(&A[0 * 3], v, &A[2 * 3]) / det;
  *c = det3x3(&A[0 * 3], &A[1 * 3], v) / det;
}


double intersect_poly2(slint_t n, double *x, double *y0, double *y1, double xmin, double xmax)
{
  double a, a2, b, b2, c, c2, p, q, r1, r2;

  interpolate_poly2(n, x, y0, &a, &b, &c);
  interpolate_poly2(n, x, y1, &a2, &b2, &c2);

  Z_ERROR_IF(TC_TRACE_IF, "poly2 0: %e  %e  %e", a, b, c);
  Z_ERROR_IF(TC_TRACE_IF, "poly2 1: %e  %e  %e", a2, b2, c2);

  p = (b - b2) / (a - a2);
  q = (c - c2) / (a - a2);

  r1 = -p / 2 + sqrt(p * p / 4 - q);
  r2 = -p / 2 - sqrt(p * p / 4 - q);

  Z_ERROR_IF(TC_TRACE_IF, "intersection: %e  %e", r1, r2);

  if (xmin <= r1 && r1 <= xmax) return r1;
  
  if (xmin <= r2 && r2 <= xmax) return r1;

  if (r1 > 0 || r2 > 0) return xmax;

  return xmin;
}
